@@ -8,8 +8,8 @@ import gettext
 _ = gettext.gettext
 
 from constants import contentModelFlags, spaceCharacters
-from constants import entitiesWindows1252, entities, voidElements
-from constants import asciiLowercase, asciiUppercase, asciiLetters
+from constants import entitiesWindows1252, entities
+from constants import asciiLowercase, asciiLetters
 from constants import digits, hexDigits, EOF
 
 from inputstream import HTMLInputStream
@@ -32,8 +32,8 @@ class HTMLTokenizer(object):
 
     # XXX need to fix documentation
 
-    def __init__(self, stream):
-        self.stream = HTMLInputStream(stream)
+    def __init__(self, stream, encoding=None):
+        self.stream = HTMLInputStream(stream, encoding)
 
         self.states = {
             "data":self.dataState,
@@ -87,19 +87,15 @@ class HTMLTokenizer(object):
 
     # Below are various helper functions the tokenizer states use worked out.
     def processSolidusInTag(self):
-        """When a solidus (/) is encountered within a tag name what happens
-        depends on whether the current tag name matches that of a void element.
-        If it matches a void element atheists did the wrong thing and if it
-        doesn't it's wrong for everyone.
+        """If the next character is a '>', convert the currentToken into
+        an EmptyTag
         """
 
-        # We need to consume another character to make sure it's a ">" before
-        # throwing an atheist parse error.
+        # We need to consume another character to make sure it's a ">"
         data = self.stream.char()
 
-        if self.currentToken["name"] in voidElements and data == u">":
-            self.tokenQueue.append({"type": "AtheistParseError", "data":
-              _("Solidus (/) incorrectly placed in tag (atheists only).")})
+        if self.currentToken["type"] == "StartTag" and data == u">":
+            self.currentToken["type"] = "EmptyTag"
         else:
             self.tokenQueue.append({"type": "ParseError", "data":
               _("Solidus (/) incorrectly placed in tag.")})
@@ -114,6 +110,9 @@ class HTMLTokenizer(object):
         If not present self.tokenQueue.append({"type": "ParseError"}) is invoked.
         """
 
+        # XXX More need to be done here. For instance, #13 should prolly be
+        # converted to #10 so we don't get \r (#13 is \r right?) in the DOM and
+        # such. Thoughts on this appreciated.
         allowed = digits
         radix = 10
         if isHex:
@@ -231,7 +230,7 @@ class HTMLTokenizer(object):
                 # discarded or needs to be put back.
                 if not charStack[-1] == ";":
                     self.tokenQueue.append({"type": "ParseError", "data":
-                      _("Named entity did not  ';'.")})
+                      _("Named entity didn't end with ';'.")})
                     self.stream.queue.extend(charStack[entityLength:])
             else:
                 self.tokenQueue.append({"type": "ParseError", "data":
@@ -249,57 +248,15 @@ class HTMLTokenizer(object):
             self.currentToken["data"][-1][1] += u"&"
 
     def emitCurrentToken(self):
-        """This method is a generic handler for emitting the StartTag,
-        EndTag, Comment and Doctype. It also sets the state to
-        "data" because that's what's needed after a token has been emitted.
+        """This method is a generic handler for emitting the tags. It also sets
+        the state to "data" because that's what's needed after a token has been
+        emitted.
         """
 
-        # Although isinstance() is http://www.canonical.org/~kragen/isinstance/
-        # considered harmful it should be ok here given that the classes are for
-        # internal usage.
-
-        token = self.currentToken
-        # For start tags convert attribute list into a distinct dictionary
-        if token["type"] == "StartTag":
-            # We need to remove the duplicate attributes and convert attributes
-            # to a dict so that [["x", "y"], ["x", "z"]] becomes {"x": "y"}
-
-            # AT When Python 2.4 is widespread we should use
-            # dict(reversed(token.data))
-            token["data"] = dict(token["data"][::-1])
-        # If an end tag has attributes it's a parse error and they should
-        # be removed
-        elif token["type"] == "EndTag" and token["data"]:
-            self.tokenQueue.append({"type": "ParseError", "data":
-              _("End tag contains unexpected attributes.")})
-            token["data"] = {}
-
         # Add token to the queue to be yielded
-        self.tokenQueue.append(token)
+        self.tokenQueue.append(self.currentToken)
         self.state = self.states["data"]
 
-    def emitCurrentTokenWithParseError(self, data=None):
-        # XXX if we want useful error messages we need to inline this method
-        """This method is equivalent to emitCurrentToken (well, it invokes it)
-        except that it also puts "data" back on the characters queue if a data
-        argument is provided and it throws a parse error."""
-        if data:
-            self.stream.queue.append(data)
-        self.tokenQueue.append({"type": "ParseError", "data":
-          _("XXX Something is wrong with the emitted token.")})
-        self.emitCurrentToken()
-
-    def attributeValueQuotedStateHandler(self, quoteType):
-        data = self.stream.char()
-        if data == quoteType:
-            self.state = self.states["beforeAttributeName"]
-        elif data == u"&":
-            self.processEntityInAttribute()
-        elif data == EOF:
-            self.emitCurrentTokenWithParseError(data)
-        else:
-            self.currentToken["data"][-1][1] += data + self.stream.charsUntil(\
-              (quoteType, u"&"))
 
     # Below are the various tokenizer states worked out.
 
@@ -349,7 +306,7 @@ class HTMLTokenizer(object):
                 self.state = self.states["closeTagOpen"]
             elif data in asciiLetters:
                 self.currentToken =\
-                  {"type": "StartTag", "name": data.lower(), "data": []}
+                  {"type": "StartTag", "name": data, "data": []}
                 self.state = self.states["tagName"]
             elif data == u">":
                 # XXX In theory it could be something besides a tag name. But
@@ -362,14 +319,14 @@ class HTMLTokenizer(object):
                 # XXX In theory it could be something besides a tag name. But
                 # do we really care?
                 self.tokenQueue.append({"type": "ParseError", "data":
-                  _("Expected tag name. Got '?' instead (HTML doesn't support processing instructions).")})
+                  _("Expected tag name. Got '?' instead (HTML doesn't "
+                  "support processing instructions).")})
                 self.stream.queue.append(data)
                 self.state = self.states["bogusComment"]
             else:
                 # XXX
                 self.tokenQueue.append({"type": "ParseError", "data":
                   _("Expected tag name. Got something else instead")})
-                # XXX can't we do "<" + data here?
                 self.tokenQueue.append({"type": "Characters", "data": u"<"})
                 self.stream.queue.append(data)
                 self.state = self.states["data"]
@@ -405,7 +362,7 @@ class HTMLTokenizer(object):
             # the stack.
             self.stream.queue.extend(charStack)
 
-            if self.currentToken["name"] == "".join(charStack[:-1]).lower() \
+            if self.currentToken["name"].lower() == "".join(charStack[:-1]).lower() \
               and charStack[-1] in (spaceCharacters |
               frozenset((u">", u"/", u"<", EOF))):
                 # Because the characters are correct we can safely switch to
@@ -426,7 +383,7 @@ class HTMLTokenizer(object):
             data = self.stream.char()
             if data in asciiLetters:
                 self.currentToken =\
-                  {"type": "EndTag", "name": data.lower(), "data": []}
+                  {"type": "EndTag", "name": data, "data": []}
                 self.state = self.states["tagName"]
             elif data == u">":
                 self.tokenQueue.append({"type": "ParseError", "data":
@@ -438,7 +395,7 @@ class HTMLTokenizer(object):
                 self.tokenQueue.append({"type": "Characters", "data": u"</"})
                 self.state = self.states["data"]
             else:
-                # XXX data can be '...
+                # XXX data can be _'_...
                 self.tokenQueue.append({"type": "ParseError", "data":
                   _("Expected closing tag. Unexpected character '" + data + "' found.")})
                 self.stream.queue.append(data)
@@ -449,16 +406,20 @@ class HTMLTokenizer(object):
         data = self.stream.char()
         if data in spaceCharacters:
             self.state = self.states["beforeAttributeName"]
-        elif data in asciiLowercase:
+        elif data in asciiLetters:
             self.currentToken["name"] += data +\
-              self.stream.charsUntil(asciiLowercase, True)
-        elif data in asciiUppercase:
-            self.currentToken["name"] += data.lower() +\
-              self.stream.charsUntil(asciiLetters, True).lower()
+              self.stream.charsUntil(asciiLetters, True)
         elif data == u">":
             self.emitCurrentToken()
-        elif data == u"<" or data == EOF:
-            self.emitCurrentTokenWithParseError(data)
+        elif data == u"<":
+            self.stream.queue.append(data)
+            self.tokenQueue.append({"type": "ParseError", "data":
+              _("Unexpected < character when getting the tag name.")})
+            self.emitCurrentToken()
+        elif data == EOF:
+            self.tokenQueue.append({"type": "ParseError", "data":
+              _("Unexpected end of file in the tag name.")})
+            self.emitCurrentToken()
         elif data == u"/":
             self.processSolidusInTag()
             self.state = self.states["beforeAttributeName"]
@@ -470,15 +431,22 @@ class HTMLTokenizer(object):
         data = self.stream.char()
         if data in spaceCharacters:
             self.stream.charsUntil(spaceCharacters, True)
-        elif data in asciiUppercase:
-            self.currentToken["data"].append([data.lower(), ""])
+        elif data in asciiLetters:
+            self.currentToken["data"].append([data, ""])
             self.state = self.states["attributeName"]
         elif data == u">":
             self.emitCurrentToken()
         elif data == u"/":
             self.processSolidusInTag()
-        elif data == u"<" or data == EOF:
-            self.emitCurrentTokenWithParseError(data)
+        elif data == u"<":
+            self.stream.queue.append(data)
+            self.tokenQueue.append({"type": "ParseError", "data":
+              _("Unexpected < character. Expected attribute name instead.")})
+            self.emitCurrentToken()
+        elif data == EOF:
+            self.tokenQueue.append({"type": "ParseError", "data":
+              _("Unexpected end of file. Expected attribute name instead.")})
+            self.emitCurrentToken()
         else:
             self.currentToken["data"].append([data, ""])
             self.state = self.states["attributeName"]
@@ -489,13 +457,9 @@ class HTMLTokenizer(object):
         leavingThisState = True
         if data == u"=":
             self.state = self.states["beforeAttributeValue"]
-        elif data in asciiLowercase:
+        elif data in asciiLetters:
             self.currentToken["data"][-1][0] += data +\
-              self.stream.charsUntil(asciiLowercase, True)
-            leavingThisState = False
-        elif data in asciiUppercase:
-            self.currentToken["data"][-1][0] += data.lower() +\
-              self.stream.charsUntil(asciiLetters, True).lower()
+              self.stream.charsUntil(asciiLetters, True)
             leavingThisState = False
         elif data == u">":
             # XXX If we emit here the attributes are converted to a dict
@@ -507,8 +471,16 @@ class HTMLTokenizer(object):
         elif data == u"/":
             self.processSolidusInTag()
             self.state = self.states["beforeAttributeName"]
-        elif data == u"<" or data == EOF:
-            self.emitCurrentTokenWithParseError(data)
+        elif data == u"<":
+            self.stream.queue.append(data)
+            self.tokenQueue.append({"type": "ParseError", "data":
+              _("Unexpected < character in attribute name.")})
+            self.emitCurrentToken()
+            leavingThisState = False
+        elif data == EOF:
+            self.tokenQueue.append({"type": "ParseError", "data":
+              _("Unexpected end of file in attribute name.")})
+            self.emitCurrentToken()
             leavingThisState = False
         else:
             self.currentToken["data"][-1][0] += data
@@ -535,14 +507,21 @@ class HTMLTokenizer(object):
             self.state = self.states["beforeAttributeValue"]
         elif data == u">":
             self.emitCurrentToken()
-        elif data in asciiUppercase:
-            self.currentToken["data"].append([data.lower(), ""])
+        elif data in asciiLetters:
+            self.currentToken["data"].append([data, ""])
             self.state = self.states["attributeName"]
         elif data == u"/":
             self.processSolidusInTag()
             self.state = self.states["beforeAttributeName"]
-        elif data == u"<" or data == EOF:
-            self.emitCurrentTokenWithParseError(data)
+        elif data == u"<":
+            self.stream.queue.append(data)
+            self.tokenQueue.append({"type": "ParseError", "data":
+              _("Unexpected < character. Expected = or end of tag.")})
+            self.emitCurrentToken()
+        elif data == EOF:
+            self.tokenQueue.append({"type": "ParseError", "data":
+              _("Unexpected end of file. Expected = or end of tag.")})
+            self.emitCurrentToken()
         else:
             self.currentToken["data"].append([data, ""])
             self.state = self.states["attributeName"]
@@ -561,22 +540,48 @@ class HTMLTokenizer(object):
             self.state = self.states["attributeValueSingleQuoted"]
         elif data == u">":
             self.emitCurrentToken()
-        elif data == u"<" or data == EOF:
-            self.emitCurrentTokenWithParseError(data)
+        elif data == u"<":
+            self.stream.queue.append(data)
+            self.tokenQueue.append({"type": "ParseError", "data":
+              _("Unexpected < character. Expected attribute value.")})
+            self.emitCurrentToken()
+        elif data == EOF:
+            self.tokenQueue.append({"type": "ParseError", "data":
+              _("Unexpected end of file. Expected attribute value.")})
+            self.emitCurrentToken()
         else:
             self.currentToken["data"][-1][1] += data
             self.state = self.states["attributeValueUnQuoted"]
         return True
 
     def attributeValueDoubleQuotedState(self):
-        # AT We could also let self.attributeValueQuotedStateHandler always
-        # return true and then return that directly here. Not sure what is
-        # faster or better...
-        self.attributeValueQuotedStateHandler(u"\"")
+        data = self.stream.char()
+        if data == "\"":
+            self.state = self.states["beforeAttributeName"]
+        elif data == u"&":
+            self.processEntityInAttribute()
+        elif data == EOF:
+            self.tokenQueue.append({"type": "ParseError", "data":
+              _("Unexpected end of file in attribute value (\").")})
+            self.emitCurrentToken()
+        else:
+            self.currentToken["data"][-1][1] += data +\
+              self.stream.charsUntil(("\"", u"&"))
         return True
 
     def attributeValueSingleQuotedState(self):
-        self.attributeValueQuotedStateHandler(u"'")
+        data = self.stream.char()
+        if data == "'":
+            self.state = self.states["beforeAttributeName"]
+        elif data == u"&":
+            self.processEntityInAttribute()
+        elif data == EOF:
+            self.tokenQueue.append({"type": "ParseError", "data":
+              _("Unexpected end of file in attribute value (').")})
+            self.emitCurrentToken()
+        else:
+            self.currentToken["data"][-1][1] += data +\
+              self.stream.charsUntil(("'", u"&"))
         return True
 
     def attributeValueUnQuotedState(self):
@@ -587,8 +592,15 @@ class HTMLTokenizer(object):
             self.processEntityInAttribute()
         elif data == u">":
             self.emitCurrentToken()
-        elif data == u"<" or data == EOF:
-            self.emitCurrentTokenWithParseError(data)
+        elif data == u"<":
+            self.stream.queue.append(data)
+            self.tokenQueue.append({"type": "ParseError", "data":
+              _("Unexpected < character in attribute value.")})
+            self.emitCurrentToken()
+        elif data == EOF:
+            self.tokenQueue.append({"type": "ParseError", "data":
+              _("Unexpected end of file in attribute value.")})
+            self.emitCurrentToken()
         else:
             self.currentToken["data"][-1][1] += data + self.stream.charsUntil( \
               frozenset(("&", ">","<")) | spaceCharacters)
@@ -633,8 +645,10 @@ class HTMLTokenizer(object):
         if data == u"-":
             self.state = self.states["commentDash"]
         elif data == EOF:
-            # XXX EMIT
-            self.emitCurrentTokenWithParseError()
+            self.tokenQueue.append({"type": "ParseError", "data":
+              _("Unexpected end of file in comment.")})
+            self.tokenQueue.append(self.currentToken)
+            self.state = self.states["data"]
         else:
             self.currentToken["data"] += data + self.stream.charsUntil(u"-")
         return True
@@ -644,8 +658,10 @@ class HTMLTokenizer(object):
         if data == u"-":
             self.state = self.states["commentEnd"]
         elif data == EOF:
-            # XXX EMIT
-            self.emitCurrentTokenWithParseError()
+            self.tokenQueue.append({"type": "ParseError", "data":
+              _("Unexpected end of file in comment (-)")})
+            self.tokenQueue.append(self.currentToken)
+            self.state = self.states["data"]
         else:
             self.currentToken["data"] += u"-" + data +\
               self.stream.charsUntil(u"-")
@@ -658,15 +674,17 @@ class HTMLTokenizer(object):
     def commentEndState(self):
         data = self.stream.char()
         if data == u">":
-            # XXX EMIT
-            self.emitCurrentToken()
+            self.tokenQueue.append(self.currentToken)
+            self.state = self.states["data"]
         elif data == u"-":
             self.tokenQueue.append({"type": "ParseError", "data":
               _("Unexpected '-' after '--' found in comment.")})
             self.currentToken["data"] += data
         elif data == EOF:
-            # XXX EMIT
-            self.emitCurrentTokenWithParseError()
+            self.tokenQueue.append({"type": "ParseError", "data":
+              _("Unexpected end of file in comment (--).")})
+            self.tokenQueue.append(self.currentToken)
+            self.state = self.states["data"]
         else:
             # XXX
             self.tokenQueue.append({"type": "ParseError", "data":
@@ -696,11 +714,15 @@ class HTMLTokenizer(object):
         elif data == u">":
             # Character needs to be consumed per the specification so don't
             # invoke emitCurrentTokenWithParseError with "data" as argument.
-            # XXX EMIT
-            self.emitCurrentTokenWithParseError()
+            self.tokenQueue.append({"type": "ParseError", "data":
+              _("Unexpected > character. Expected DOCTYPE name.")})
+            self.tokenQueue.append(self.currentToken)
+            self.state = self.states["data"]
         elif data == EOF:
-            # XXX EMIT
-            self.emitCurrentTokenWithParseError()
+            self.tokenQueue.append({"type": "ParseError", "data":
+              _("Unexpected end of file. Expected DOCTYPE name.")})
+            self.tokenQueue.append(self.currentToken)
+            self.state = self.states["data"]
         else:
             self.currentToken["name"] = data
             self.state = self.states["doctypeName"]
@@ -716,8 +738,10 @@ class HTMLTokenizer(object):
             self.tokenQueue.append(self.currentToken)
             self.state = self.states["data"]
         elif data == EOF:
-            # XXX EMIT
-            self.emitCurrentTokenWithParseError()
+            self.tokenQueue.append({"type": "ParseError", "data":
+              _("Unexpected end of file in DOCTYPE name.")})
+            self.tokenQueue.append(self.currentToken)
+            self.state = self.states["data"]
         else:
             # We can't just uppercase everything that arrives here. For
             # instance, non-ASCII characters.
@@ -742,7 +766,11 @@ class HTMLTokenizer(object):
         elif data == EOF:
             self.currentToken["data"] = True
             # XXX EMIT
-            self.emitCurrentTokenWithParseError(data)
+            self.stream.queue.append(data)
+            self.tokenQueue.append({"type": "ParseError", "data":
+              _("Unexpected end of file in DOCTYPE.")})
+            self.tokenQueue.append(self.currentToken)
+            self.state = self.states["data"]
         else:
             self.tokenQueue.append({"type": "ParseError", "data":
               _("Expected space or '>'. Got '" + data + "'")})
@@ -757,7 +785,11 @@ class HTMLTokenizer(object):
             self.state = self.states["data"]
         elif data == EOF:
             # XXX EMIT
-            self.emitCurrentTokenWithParseError(data)
+            self.stream.queue.append(data)
+            self.tokenQueue.append({"type": "ParseError", "data":
+              _("Unexpected end of file in bogus doctype.")})
+            self.tokenQueue.append(self.currentToken)
+            self.state = self.states["data"]
         else:
             pass
         return True
