@@ -11,6 +11,12 @@ def buildTestSuite():
     return suite
 
 def main():
+    # the following is temporary while the unit tests for parse errors are
+    # still in flux
+    if '-p' in sys.argv: # suppress check for parse errors
+        import test_parser
+        test_parser.checkParseErrors = False
+
     unittest.TextTestRunner().run(buildTestSuite())
 
 if __name__ == "__main__":
